/*
  ecd2xcf.cc

  Copyright (C) 2013 Taiji Yamada <taiji@aihara.co.jp>
 */
#include <iostream>
#include <fstream>
#include <string>
#include <map>
#include <boost/filesystem.hpp>
#include <boost/optional.hpp>
#include "date_time_numeric.hpp"
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#if defined(USE_BOOST_XPRESSIVE)
#include "boost_xpressive_utility.hpp"
using namespace boost::xpressive;
typedef string_sregex_map string_regex_map;
#else
#include "boost_regex_utility.hpp"
#endif
#include "repeater.hpp"
#include "date_time_repeater.hpp"
using namespace std;
using namespace boost;

#define RE_YEAR_NUMS "[0-9]{4}"
#define RE_MONTH_NAMS "(?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)[a-z]*"
#define RE_MONTH_NUMS "0?[1-9]|1[0-2]"
#define RE_DATE_NUMS "0?[1-9]|[12][0-9]|3[0-1]"
#define RE_WEEK_NAMS "(?:Su|Mo|Tu|We|Th|Fr|Sa)[a-z]*"

int usage(ostream &os, int argc, char *argv[])
{
  (void)argc;
  os << "usage:\t" << argv[0] << " [--help|-h] -d ~/Calendar/ [--since|until date] [--] -|filename.ecd [...]" << endl;
  return os == cerr ? 1 : 0;
}

int main(int argc, char* argv[])
{
  filesystem::path dir;
  struct vevent_strings {
    string summary, categories;
    string rrule;
    posix_time::ptime dt0, dt1;
  } defaults[2] = { { "", }, { "", } };
  const gregorian::date temporary_dt(1995, 1, 1);
  string temporary_year(lexical_cast<string>(temporary_dt.year_month_day().year));

  enum ecd_type {
    ecd_diary,
    ecd_weekly,
    ecd_monthly,
    ecd_anniversary,
    ecd_block,
    ecd_cyclic,
  } ecd_kind = ecd_diary;
  streambuf *cin_rdbuf = cin.rdbuf();
  pair<boost::optional<posix_time::ptime>, boost::optional<posix_time::ptime> > date_time_range;
  vector<string> filenames;
  map<string, bool> append_map;
  streambuf *cout_rdbuf = cout.rdbuf();
  int a;
  for (a=1; a<argc; ++a)
    if (string(argv[a]) == "--" && ++a)
      break;
    else if (string(argv[a]) == "--help" || string(argv[a]) == "-h")
      return usage(cout, argc, argv);
    else if (string(argv[a]) == "-d" && a+1 < argc)
      dir = argv[++a];
    else if (string(argv[a]) == "--since" && a+1 < argc)
      date_time_range.first = boost::optional<posix_time::ptime>(posix_time::ptime(gregorian::from_undelimited_string(argv[++a]), posix_time::time_duration()));
    else if (string(argv[a]) == "--until" && a+1 < argc)
      date_time_range.second = boost::optional<posix_time::ptime>(posix_time::ptime(gregorian::from_undelimited_string(argv[++a]), posix_time::time_duration(23, 59, 59)));
    else
      filenames.push_back(argv[a]);

  if (!filenames.size() || !(filesystem::exists(dir) && filesystem::is_directory(dir)))
    return usage(cerr, argc, argv);

  string_regex_map regex_map;
  for (size_t i=0; i<filenames.size(); ++i) {
    ifstream ifs(filenames[i].c_str());
    cin.rdbuf((filenames[i] != "-") ? ifs.rdbuf() : cin_rdbuf);
    string sb;
    while (getline(cin, sb)) {
      smatch m;
      if (regex_match(sb, m, regex_map.at("^&?("RE_MONTH_NAMS"|"RE_MONTH_NUMS")\\s+("RE_DATE_NUMS")(?:,?\\s+("RE_YEAR_NUMS"))?(?:\\s+(.+))?$", regex_constants::icase))) {
	defaults[1] = defaults[0];
	{
	  string dtstart;
	  if (!m[1].str().empty() && !m[2].str().empty()) dtstart = (m[3].str().empty() ? temporary_year : m[3].str()) + "/" + m[1].str() + "/" + m[2].str();
	  if (!m[4].str().empty()) defaults[1].summary = ':' + m[4].str();
	  {
	    string date0;
	    date0 = to_iso_string(gregorian::from_string(dtstart));
	    defaults[1].dt0 = posix_time::ptime(gregorian::from_undelimited_string(date0));
	    defaults[1].dt1 = posix_time::ptime(gregorian::from_undelimited_string(date0) + gregorian::date_duration(1));
	    if (m[3].str().empty()) defaults[1].rrule = ":FREQ=YEARLY";
	  }
	  ecd_kind = ecd_diary;
	  if (defaults[1].summary.empty()) continue;
	}
      }
      else if (regex_match(sb, m, regex_map.at("^&?("RE_WEEK_NAMS")(?:,?\\s+("RE_YEAR_NUMS"))?(?:\\s+(.+))?$", regex_constants::icase))) {
	defaults[1] = defaults[0];
	{
	  string byday;
	  if (!m[1].str().empty()) byday = m[1].str().substr(0, 2);
	  if (!m[3].str().empty()) defaults[1].summary = ':' + m[3].str();
	  {
	    defaults[1].dt0 = posix_time::ptime(gregorian::from_undelimited_string((m[2].str().empty() ? temporary_year : m[2].str()) + "0101"));
	    defaults[1].dt1 = posix_time::ptime(gregorian::from_undelimited_string((m[2].str().empty() ? temporary_year : m[2].str()) + "0101") + gregorian::date_duration(1));
	    defaults[1].rrule = ":FREQ=WEEKLY;BYDAY=" + byday;
	  }
	  ecd_kind = ecd_weekly;
	  if (defaults[1].summary.empty()) continue;
	}
      }
      else if (regex_match(sb, m, regex_map.at("^&?\\*\\s+("RE_DATE_NUMS")(?:,?\\s+("RE_YEAR_NUMS"))?(?:\\s+(.+))?$", regex_constants::icase))) {
	defaults[1] = defaults[0];
	{
	  string dtstart;
	  if (!m[1].str().empty()) dtstart = (m[2].str().empty() ? temporary_year : m[2].str()) + "/" + "01" + "/" + m[1].str();
	  if (!m[3].str().empty()) defaults[1].summary = ':' + m[3].str();
	  {
	    string date0;
	    date0 = to_iso_string(gregorian::from_string(dtstart));
	    defaults[1].dt0 = posix_time::ptime(gregorian::from_undelimited_string(date0));
	    defaults[1].dt1 = posix_time::ptime(gregorian::from_undelimited_string(date0) + gregorian::date_duration(1));
	    defaults[1].rrule = ":FREQ=MONTHLY";
	    if (!m[2].str().empty()) defaults[1].rrule += (format(";UNTIL=%1%T000000Z") % (m[2] + "12" + "31")).str();
	  }
	  ecd_kind = ecd_monthly;
	  if (defaults[1].summary.empty()) continue;
	}
      }
      else if (regex_match(sb, m, regex_map.at("^&?%%\\(diary-anniversary\\s+("RE_MONTH_NAMS"|"RE_MONTH_NUMS")\\s+("RE_DATE_NUMS")(?:,?\\s+("RE_YEAR_NUMS"))?\\)(?:\\s+(.+))?$", regex_constants::icase))) {
	defaults[1] = defaults[0];
	{
	  string dtstart;
	  if (!m[1].str().empty() && !m[2].str().empty()) dtstart = (m[3].str().empty() ? temporary_year : m[3].str()) + "/" + m[1].str() + "/" + m[2].str();
	  if (!m[4].str().empty()) defaults[1].summary = ':' + m[4].str();
	  {
	    string date0;
	    date0 = to_iso_string(gregorian::from_string(dtstart));
	    defaults[1].dt0 = posix_time::ptime(gregorian::from_undelimited_string(date0));
	    defaults[1].dt1 = posix_time::ptime(gregorian::from_undelimited_string(date0) + gregorian::date_duration(1));
	    defaults[1].rrule = ":FREQ=YEARLY";
	    defaults[1].categories = ":ANNIVERSARY";
	  }
	  ecd_kind = ecd_anniversary;
	  if (defaults[1].summary.empty()) continue;
	}
      }
      else if (regex_match(sb, m, regex_map.at("^&?%%\\(diary-block\\s+("RE_MONTH_NAMS"|"RE_MONTH_NUMS")\\s+("RE_DATE_NUMS")(?:,?\\s+("RE_YEAR_NUMS"))\\s+("RE_MONTH_NAMS"|"RE_MONTH_NUMS")\\s+("RE_DATE_NUMS")(?:,?\\s+("RE_YEAR_NUMS"))\\)(?:\\s+(.+))?$", regex_constants::icase))) {
	defaults[1] = defaults[0];
	{
	  string dtstart, dtend;
	  if (!m[1].str().empty() && !m[2].str().empty() && !m[3].str().empty()) dtstart = m[3].str() + "/" + m[1].str() + "/" + m[2].str();
	  if (!m[4].str().empty() && !m[5].str().empty() && !m[6].str().empty()) dtend   = m[6].str() + "/" + m[4].str() + "/" + m[5].str();
	  if (!m[7].str().empty()) defaults[1].summary = ':' + m[7].str();
	  {
	    string date0, date1;
	    date0 = to_iso_string(gregorian::from_string(dtstart));
	    date1 = to_iso_string(gregorian::from_string(dtend));
	    defaults[1].dt0 = posix_time::ptime(gregorian::from_undelimited_string(date0));
	    defaults[1].dt1 = posix_time::ptime(gregorian::from_undelimited_string(date1) + gregorian::date_duration(1));
	  }
	  ecd_kind = ecd_block;
	  if (defaults[1].summary.empty()) continue;
	}
      }
      else if (regex_match(sb, m, regex_map.at("^&?%%\\(diary-cyclic\\s+(\\d+)\\s+("RE_MONTH_NAMS"|"RE_MONTH_NUMS")\\s+("RE_DATE_NUMS")(?:,?\\s+("RE_YEAR_NUMS"))?\\)(?:\\s+(.+))?$", regex_constants::icase))) {
	defaults[1] = defaults[0];
	{
	  string dtstart, interval;
	  if (!m[1].str().empty()) interval = m[1].str();
	  if (!m[2].str().empty() && !m[3].str().empty()) dtstart = (m[4].str().empty() ? temporary_year : m[4].str()) + "/" + m[2].str() + "/" + m[3].str();
	  if (!m[5].str().empty()) defaults[1].summary = ':' + m[5].str();
	  {
	    string date0;
	    date0 = to_iso_string(gregorian::from_string(dtstart));
	    defaults[1].dt0 = posix_time::ptime(gregorian::from_undelimited_string(date0));
	    defaults[1].dt1 = posix_time::ptime(gregorian::from_undelimited_string(date0) + gregorian::date_duration(1));
	    defaults[1].rrule = ":FREQ=DAILY;INTERVAL=" + interval;
	  }
	  ecd_kind = ecd_cyclic;
	  if (defaults[1].summary.empty()) continue;
	}
      }
      else if (!regex_match(sb, m, regex_map.at("^\\s*$")) && regex_match(sb, m, regex_map.at("^\\s+(.*)$")))
	defaults[1].summary = ':' + m[1].str();
      else
	continue;
      switch (ecd_kind) {
      case ecd_diary:
      case ecd_weekly:
      case ecd_monthly:
      case ecd_anniversary:
      case ecd_block:
      case ecd_cyclic:
	{
	  struct vevent_strings current = defaults[1];
	  if (!current.rrule.empty()) {
	    cerr << "unsupported recurrence rule at " << current.dt0.date() << " for " << current.summary.substr(1) << endl;
	    continue;
	  }
	  if (date_time_range.first  && date_time_range.first->date()  > current.dt0.date() ||
	      date_time_range.second && date_time_range.second->date() < current.dt0.date()) continue;
	  string filename((format("xc%3$d%2$s%1$04d") % current.dt0.date().year() % current.dt0.date().month() % current.dt0.date().day()).str());
	  ofstream ofs((dir/filename).c_str(), append_map[filename] ? std::ios::app : std::ios::out);
	  append_map[filename] = true;
	  cout.rdbuf(ofs.rdbuf());
	  cout << current.summary.substr(1) << endl;
	  ofs.close();
	}
	break;
      }
    }
  }
  cout.rdbuf(cout_rdbuf);
  return 0;
}
