/*
  ics2ecd.cc

  Copyright (C) 2013 Taiji Yamada <taiji@aihara.co.jp>
 */
#include <iostream>
#include <fstream>
#include <string>
#include <boost/optional.hpp>
#include "date_time_numeric.hpp"
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#if defined(USE_BOOST_XPRESSIVE)
#include "boost_xpressive_utility.hpp"
using namespace boost::xpressive;
typedef string_sregex_map string_regex_map;
#else
#include "boost_regex_utility.hpp"
#endif
using namespace std;
using namespace boost;

int usage(ostream &os, int argc, char *argv[])
{
  (void)argc;
  os << "usage:\t" << argv[0] << " [--help|-h] [--since|until date] [--] -|filename.ics [...]" << endl;
  return os == cerr ? 1 : 0;
}

int main(int argc, char *argv[])
{
  streambuf *cin_rdbuf = cin.rdbuf();
  pair<boost::optional<posix_time::ptime>, boost::optional<posix_time::ptime> > date_time_range;
  vector<string> filenames;
  int a;
  for (a=1; a<argc; ++a)
    if (string(argv[a]) == "--" && ++a)
      break;
    else if (string(argv[a]) == "--help" || string(argv[a]) == "-h")
      return usage(cout, argc, argv);
    else if (string(argv[a]) == "--since" && a+1 < argc)
      date_time_range.first = boost::optional<posix_time::ptime>(posix_time::ptime(gregorian::from_undelimited_string(argv[++a]), posix_time::time_duration()));
    else if (string(argv[a]) == "--until" && a+1 < argc)
      date_time_range.second = boost::optional<posix_time::ptime>(posix_time::ptime(gregorian::from_undelimited_string(argv[++a]), posix_time::time_duration(23, 59, 59)));
    else
      filenames.push_back(argv[a]);

  if (!filenames.size())
    return usage(cerr, argc, argv);

  struct {
    string summary, description, location, categories, transp;
    posix_time::ptime dtstart, dtend;
    string rdate, rrule/*, exrule, related_to*/;
    vector<string> exdate;
  } vevent, vevent0;
  map<string, int> prologue_state_count, epilogue_state_count;
  string_regex_map regex_map;
  for (size_t i=0; i<filenames.size(); ++i) {
    vector<string> state;
    ifstream ifs(filenames[i].c_str());
    cin.rdbuf((filenames[i] != "-") ? ifs.rdbuf() : cin_rdbuf);
    string sb;
    while (getline(cin, sb)) {
      if (sb.size() && sb[sb.size()-1] == '\r') sb.resize(sb.size()-1);
      while (!cin.eof()) {
	char c = cin.get();
	cin.putback(c);
	if (c == ' ') {
	  string cb;
	  getline(cin, cb);
	  if (cb.size() && cb[cb.size()-1] == '\r') cb.resize(cb.size()-1);
	  sb += "\r\n" + cb;
	}
	else
	  break;
      }
      string str(sb);
      struct {
	bool search_result;
	smatch match;
      } prologue, epilogue;
      prologue.search_result =
	regex_match(str, prologue.match, regex_map["^BEGIN:(VCALENDAR|VEVENT|VALARM|VTODO)$"]);
      epilogue.search_result =
	regex_match(str, epilogue.match, regex_map["^END:(VCALENDAR|VEVENT|VALARM|VTODO)$"]);
      if (prologue.search_result) {
	state.push_back(prologue.match[1].str());
	++prologue_state_count[state.back()];
      }

      if (state.back() == "VCALENDAR" && prologue_state_count[state.back()] == 1 &&
	  epilogue_state_count["VEVENT"] == 1)
	;
      else if (state.back() == "VCALENDAR" && prologue_state_count[state.back()] == 1)
	;
      else if (state.back() == "VEVENT") {
	smatch m;
	sb = regex_replace(sb, regex_map["\\t"], "\\\\t");
	sb = regex_replace(sb, regex_map["\\r\\n"],
#if defined(USE_BOOST_XPRESSIVE)
			   "\\r\\n"
#else
			   "\\\\r\\\\n"
#endif
			   );
	if (regex_match(sb, m, regex_map["^SUMMARY(.*)$"])) {
	  vevent.summary = m[1];
	}
	else if (regex_match(sb, m, regex_map["^DESCRIPTION(.*)$"])) {
	  vevent.description = m[1];
	}
	else if (regex_match(sb, m, regex_map["^LOCATION(.*)$"])) {
	  vevent.location = m[1];
	}
	else if (regex_match(sb, m, regex_map["^CATEGORIES(.*)$"])) {
	  vevent.categories = m[1];
	}
	else if (regex_match(sb, m, regex_map["^DTSTART;VALUE=DATE:(.*)$"])) {
	  vevent.dtstart = posix_time::ptime(gregorian::from_undelimited_string(m[1]));
	}
	else if (regex_match(sb, m, regex_map["^DTSTART[^:]*:(.*)$"])) {
	  vevent.dtstart = posix_time::from_iso_string(m[1]);
	}
	else if (regex_match(sb, m, regex_map["^DTEND;VALUE=DATE:(.*)$"])) {
	  vevent.dtend = posix_time::ptime(gregorian::from_undelimited_string(m[1]));
	}
	else if (regex_match(sb, m, regex_map["^DTEND[^:]*:(.*)$"])) {
	  vevent.dtend = posix_time::from_iso_string(m[1]);
	}
	else if (regex_match(sb, m, regex_map["^TRANSP(.*)$"])) {
	  vevent.transp = m[1];
	}
	else if (regex_match(sb, m, regex_map["^RRULE(.*)$"])) {
	  vevent.rrule = m[1];
	}
	else if (regex_match(sb, m, regex_map["^RDATE(.*)$"])) {
	  vevent.rdate = m[1];
	}
	else if (regex_match(sb, m, regex_map["^EXDATE(.*)$"])) {
	  cerr << format("unsupported EXDATE%1%\n") % m[1];
	}
      }

      if (epilogue.search_result && epilogue.match[1].str() == "VEVENT") {
	if (date_time_range.first  && date_time_range.first->date()  > vevent.dtstart.date() ||
	    date_time_range.second && date_time_range.second->date() < vevent.dtstart.date()) {
	  vevent = vevent0;
	  continue;
	}
	smatch m;
	if (!epilogue_state_count["VEVENT"])
	  {}
	if (!(vevent.dtstart.time_of_day() == posix_time::time_duration() && vevent.dtend.time_of_day() == posix_time::time_duration())) {
	  cerr << format("illegal DTSTART|DTEND\n");
	}
	if (vevent.rrule.size() > 1 && regex_match(vevent.rrule, m, regex_map["^:FREQ=DAILY;INTERVAL=(\\d+)$"])) {
	  cout << format("%%%%(diary-cyclic %1% %2%)") % m[1] %
	    (format("%2% %3% %1%") % vevent.dtstart.date().year() % static_cast<int>(vevent.dtstart.date().month()) % vevent.dtstart.date().day());
	  vevent.rrule.clear();
	}
	else if (vevent.rrule.size() > 1 && regex_match(vevent.rrule, m, regex_map["^:FREQ=MONTHLY;UNTIL=(\\d{4})1231.*$"])) {
	  cout << format("* %1%, %2%") % vevent.dtstart.date().day() % vevent.dtstart.date().year();
	  vevent.rrule.clear();
	}
	else if (vevent.rrule.size() > 1 && regex_match(vevent.rrule, m, regex_map["^:FREQ=MONTHLY$"])) {
	  cout << format("* %1%") % vevent.dtstart.date().day();
	  vevent.rrule.clear();
	}
	else if (vevent.rrule.size() > 1 && regex_match(vevent.rrule, m, regex_map["^:FREQ=WEEKLY;BYDAY=(\\w+)$"])) {
	  cout << format("%1%") %
	    (m[1] == "Su" ? "Sunday" :
	     m[1] == "Mo" ? "Monday" :
	     m[1] == "Tu" ? "Tuesday" :
	     m[1] == "We" ? "Wednesday" :
	     m[1] == "Th" ? "Thursday" :
	     m[1] == "Fr" ? "Friday" :
	     m[1] == "Sa" ? "Saturday" :
	     m[1].str());
	  vevent.rrule.clear();
	}
	else if (vevent.rrule.size() > 1 && regex_match(vevent.rrule, m, regex_map["^:FREQ=YEARLY$"])) {
	  if (vevent.categories == ":ANNIVERSARY") {
	    cout << format("%%%%(diary-anniversary %1%)") %
	      (format("%2% %3% %1%") % vevent.dtstart.date().year() % static_cast<int>(vevent.dtstart.date().month()) % vevent.dtstart.date().day());
	    vevent.categories.clear();
	  }
	  else
	    cout << format("%1% %2%") % vevent.dtstart.date().month() % vevent.dtstart.date().day();
	  vevent.rrule.clear();
	}
	else {
	  if (vevent.dtstart + gregorian::days(1) != vevent.dtend) {
	    vevent.dtend -= gregorian::days(1);
	    cout << format("%%%%(diary-block %1% %2%)") %
	      (format("%2% %3% %1%") % vevent.dtstart.date().year() % static_cast<int>(vevent.dtstart.date().month()) % vevent.dtstart.date().day()) %
	      (format("%2% %3% %1%") % vevent.dtend.date().year() % static_cast<int>(vevent.dtend.date().month()) % vevent.dtend.date().day());
	  }
	  else
	    cout << format("%2% %3%, %1%") % vevent.dtstart.date().year() % vevent.dtstart.date().month() % vevent.dtstart.date().day();
	}
	if (vevent.rdate.size() > 1)		cerr << format("illegal RDATE%1%\n") % vevent.rdate;
	if (vevent.rrule.size() > 1)		cerr << format("illegal RRULE%1%\n") % vevent.rrule;
	if (vevent.summary.size() > 1)		cout << format("\t%1%")	% vevent.summary.substr(1);
	if (vevent.description.size() > 1)	cout << format("(%1%)")	% vevent.description.substr(1);
	if (vevent.location.size() > 1)		cout << format("{%1%}")	% vevent.location.substr(1);
	if (vevent.categories.size() > 1)	cout << format("[%1%]")	% vevent.categories;
	if (vevent.transp.size() > 1)		if (vevent.transp.substr(1) != "TRANSPARENT") cerr << format("illegal TRANSP%1%\n") % vevent.transp;
	cout << endl;
	vevent = vevent0;
      }

      if (epilogue.search_result) {
	++epilogue_state_count[state.back()];
	state.pop_back();
      }
    }
  }
  return 0;
}
