/*
  xcf2ecd.cc

  Copyright (C) 2013 Taiji Yamada <taiji@aihara.co.jp>
 */
#include <iostream>
#include <fstream>
#include <string>
#include <boost/filesystem.hpp>
#include <boost/optional.hpp>
#include "date_time_numeric.hpp"
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#if defined(USE_BOOST_XPRESSIVE)
#include "boost_xpressive_utility.hpp"
using namespace boost::xpressive;
typedef string_sregex_map string_regex_map;
#else
#include "boost_regex_utility.hpp"
#endif
#include "repeater.hpp"
#include "date_time_repeater.hpp"
using namespace std;
using namespace boost;

#define RE_YEAR_NUMS "[0-9]{4}"
#define RE_MONTH_NAMS "(?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)[a-z]*"
#define RE_MONTH_NUMS "0?[1-9]|1[0-2]"
#define RE_DATE_NUMS "0?[1-9]|[12][0-9]|3[0-1]"
#define RE_WEEK_NAMS "(?:Su|Mo|Tu|We|Th|Fr|Sa)[a-z]*"

int usage(ostream &os, int argc, char *argv[])
{
  (void)argc;
  os << "usage:\t" << argv[0] << " [--help|-h] [--concat_lines] [--since|until date] [--] ~/Calendar/ [...]" << endl;
  return os == cerr ? 1 : 0;
}

int main(int argc, char* argv[])
{
  bool concat_lines = false;
  streambuf *cin_rdbuf = cin.rdbuf();
  pair<boost::optional<posix_time::ptime>, boost::optional<posix_time::ptime> > date_time_range;
  int a;
  for (a=1; a<argc; ++a)
    if (string(argv[a]) == "--" && ++a)
      break;
    else if (string(argv[a]) == "--help" || string(argv[a]) == "-h")
      return usage(cout, argc, argv);
    else if (string(argv[a]) == "--concat_lines")
      concat_lines = true;
    else if (string(argv[a]) == "--since" && a+1 < argc)
      date_time_range.first = boost::optional<posix_time::ptime>(posix_time::ptime(gregorian::from_undelimited_string(argv[++a]), posix_time::time_duration()));
    else if (string(argv[a]) == "--until" && a+1 < argc)
      date_time_range.second = boost::optional<posix_time::ptime>(posix_time::ptime(gregorian::from_undelimited_string(argv[++a]), posix_time::time_duration(23, 59, 59)));
    else
      break;

  if (!(a<argc))
    return usage(cerr, argc, argv);

  string_regex_map regex_map;
  for (; a<argc; ++a) {
    filesystem::path p(argv[a]);
    if (filesystem::exists(p) && filesystem::is_directory(p)) {
      vector<filesystem::path> ls;
      copy(filesystem::directory_iterator(p), filesystem::directory_iterator(), back_inserter(ls));
      for (vector<filesystem::path>::const_iterator it = ls.begin(); it != ls.end(); ++it) {
	smatch m;
	if (!regex_match(it->filename().string(), m, regex_map.at("^xc("RE_DATE_NUMS")("RE_MONTH_NAMS")("RE_YEAR_NUMS")$", regex_constants::icase)))
	  continue;
	struct { string y, m, d; } ymd = { m[3].str(), m[2].str(), m[1].str(), };
	string ds((format("%1%/%2%/%3%") % ymd.y % ymd.m % ymd.d).str());
	gregorian::date d(gregorian::from_string(ds));
	if (date_time_range.first  && date_time_range.first->date()  > d ||
	    date_time_range.second && date_time_range.second->date() < d) continue;
	ifstream ifs(it->c_str());
	cin.rdbuf(ifs.rdbuf());
	string sb, buf;
	if (!concat_lines) {
	  while (getline(cin, sb))
	    if (!regex_match(sb, m, regex_map.at("^\\s*$")))
	      buf += '\t' + sb + '\n';
	  if (!buf.empty())
	    cout << format("%2% %3%, %1%\n%4%") % d.year() % d.month() % d.day() % buf;
	}
	else {
	  while (getline(cin, sb))
	    buf += (!buf.empty() ? " " : "") + sb;
	  if (!buf.empty())
	    cout << format("%2% %3%, %1%\t%4%\n") % d.year() % d.month() % d.day() % buf;
	}
      }
    }
  }
  cin.rdbuf(cin_rdbuf);
  return 0;
}
